from __future__ import annotations

import json
import logging
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Collection, Iterator, Optional, Type

from slixmpp import JID, Iq, Message, Presence
from slixmpp.exceptions import XMPPError
from sqlalchemy import Engine, delete, select, update
from sqlalchemy.orm import Session, attributes
from sqlalchemy.sql.functions import count

from ..util.archive_msg import HistoryMessage
from ..util.types import URL, CachedPresence
from ..util.types import Hat as HatTuple
from ..util.types import MamMetadata, MucAffiliation, MucRole
from .meta import Base
from .models import (
    ArchivedMessage,
    ArchivedMessageSource,
    Attachment,
    Avatar,
    Contact,
    ContactSent,
    GatewayUser,
    Hat,
    LegacyIdsMulti,
    Participant,
    Room,
    XmppIdsMulti,
    XmppToLegacyEnum,
    XmppToLegacyIds,
)

if TYPE_CHECKING:
    from ..contact.contact import LegacyContact
    from ..group.participant import LegacyParticipant
    from ..group.room import LegacyMUC


class EngineMixin:
    def __init__(self, engine: Engine):
        self._engine = engine

    @contextmanager
    def session(self, **session_kwargs) -> Iterator[Session]:
        global _session
        if _session is not None:
            yield _session
            return
        with Session(self._engine, **session_kwargs) as session:
            _session = session
            try:
                yield session
            finally:
                _session = None


class UpdatedMixin(EngineMixin):
    model: Type[Base] = NotImplemented

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        with self.session() as session:
            session.execute(
                delete(self.model).where(~self.model.updated)  # type:ignore
            )
            session.execute(update(self.model).values(updated=False))  # type:ignore
            session.commit()

    def get_by_pk(self, pk: int) -> Optional[Base]:
        with self.session() as session:
            return session.execute(
                select(self.model).where(self.model.id == pk)  # type:ignore
            ).scalar()


class SlidgeStore(EngineMixin):
    def __init__(self, engine: Engine):
        super().__init__(engine)
        self.users = UserStore(engine)
        self.avatars = AvatarStore(engine)
        self.contacts = ContactStore(engine)
        self.mam = MAMStore(engine)
        self.multi = MultiStore(engine)
        self.attachments = AttachmentStore(engine)
        self.rooms = RoomStore(engine)
        self.sent = SentStore(engine)
        self.participants = ParticipantStore(engine)


class UserStore(EngineMixin):
    def new(self, jid: JID, legacy_module_data: dict) -> GatewayUser:
        if jid.resource:
            jid = JID(jid.bare)
        with self.session(expire_on_commit=False) as session:
            user = session.execute(
                select(GatewayUser).where(GatewayUser.jid == jid)
            ).scalar()
            if user is not None:
                return user
            user = GatewayUser(jid=jid, legacy_module_data=legacy_module_data)
            session.add(user)
            session.commit()
        return user

    def update(self, user: GatewayUser):
        # https://github.com/sqlalchemy/sqlalchemy/discussions/6473
        attributes.flag_modified(user, "legacy_module_data")
        attributes.flag_modified(user, "preferences")
        with self.session() as session:
            session.add(user)
            session.commit()

    def get_all(self) -> Iterator[GatewayUser]:
        with self.session() as session:
            yield from session.execute(select(GatewayUser)).scalars()

    def get(self, jid: JID) -> Optional[GatewayUser]:
        with self.session() as session:
            return session.execute(
                select(GatewayUser).where(GatewayUser.jid == jid.bare)
            ).scalar()

    def get_by_stanza(self, stanza: Iq | Message | Presence) -> Optional[GatewayUser]:
        return self.get(stanza.get_from())

    def delete(self, jid: JID) -> None:
        with self.session() as session:
            session.delete(self.get(jid))
            session.commit()


class AvatarStore(EngineMixin):
    def get_by_url(self, url: URL | str) -> Optional[Avatar]:
        with self.session() as session:
            return session.execute(select(Avatar).where(Avatar.url == url)).scalar()

    def get_by_legacy_id(self, legacy_id: str) -> Optional[Avatar]:
        with self.session() as session:
            return session.execute(
                select(Avatar).where(Avatar.legacy_id == legacy_id)
            ).scalar()

    def get_by_pk(self, pk: int) -> Optional[Avatar]:
        with self.session() as session:
            return session.execute(select(Avatar).where(Avatar.id == pk)).scalar()

    def delete_by_pk(self, pk: int):
        with self.session() as session:
            session.execute(delete(Avatar).where(Avatar.id == pk))
            session.commit()

    def get_all(self) -> Iterator[Avatar]:
        with self.session() as session:
            yield from session.execute(select(Avatar)).scalars()


class SentStore(EngineMixin):
    def set_message(self, user_pk: int, legacy_id: str, xmpp_id: str) -> None:
        with self.session() as session:
            msg = XmppToLegacyIds(
                user_account_id=user_pk,
                legacy_id=legacy_id,
                xmpp_id=xmpp_id,
                type=XmppToLegacyEnum.DM,
            )
            session.add(msg)
            session.commit()

    def get_xmpp_id(self, user_pk: int, legacy_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.xmpp_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.legacy_id == legacy_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.DM)
            ).scalar()

    def get_legacy_id(self, user_pk: int, xmpp_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.legacy_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.xmpp_id == xmpp_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.DM)
            ).scalar()

    def set_group_message(self, user_pk: int, legacy_id: str, xmpp_id: str) -> None:
        with self.session() as session:
            msg = XmppToLegacyIds(
                user_account_id=user_pk,
                legacy_id=legacy_id,
                xmpp_id=xmpp_id,
                type=XmppToLegacyEnum.GROUP_CHAT,
            )
            session.add(msg)
            session.commit()

    def get_group_xmpp_id(self, user_pk: int, legacy_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.xmpp_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.legacy_id == legacy_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.GROUP_CHAT)
            ).scalar()

    def get_group_legacy_id(self, user_pk: int, xmpp_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.legacy_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.xmpp_id == xmpp_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.GROUP_CHAT)
            ).scalar()

    def set_thread(self, user_pk: int, legacy_id: str, xmpp_id: str) -> None:
        with self.session() as session:
            msg = XmppToLegacyIds(
                user_account_id=user_pk,
                legacy_id=legacy_id,
                xmpp_id=xmpp_id,
                type=XmppToLegacyEnum.THREAD,
            )
            session.add(msg)
            session.commit()

    def get_legacy_thread(self, user_pk: int, xmpp_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.legacy_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.xmpp_id == xmpp_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.THREAD)
            ).scalar()

    def get_xmpp_thread(self, user_pk: int, legacy_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(XmppToLegacyIds.xmpp_id)
                .where(XmppToLegacyIds.user_account_id == user_pk)
                .where(XmppToLegacyIds.legacy_id == legacy_id)
                .where(XmppToLegacyIds.type == XmppToLegacyEnum.THREAD)
            ).scalar()

    def was_sent_by_user(self, user_pk: int, legacy_id: str) -> bool:
        with self.session() as session:
            return (
                session.execute(
                    select(XmppToLegacyIds.legacy_id)
                    .where(XmppToLegacyIds.user_account_id == user_pk)
                    .where(XmppToLegacyIds.legacy_id == legacy_id)
                ).scalar()
                is not None
            )


class ContactStore(UpdatedMixin):
    model = Contact

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        with self.session() as session:
            session.execute(update(Contact).values(cached_presence=False))
            session.commit()

    def get_all(self, user_pk: int) -> Iterator[Contact]:
        with self.session() as session:
            yield from session.execute(
                select(Contact).where(Contact.user_account_id == user_pk)
            ).scalars()

    def get_by_jid(self, user_pk: int, jid: JID) -> Optional[Contact]:
        with self.session() as session:
            return session.execute(
                select(Contact)
                .where(Contact.jid == jid.bare)
                .where(Contact.user_account_id == user_pk)
            ).scalar()

    def get_by_legacy_id(self, user_pk: int, legacy_id: str) -> Optional[Contact]:
        with self.session() as session:
            return session.execute(
                select(Contact)
                .where(Contact.legacy_id == legacy_id)
                .where(Contact.user_account_id == user_pk)
            ).scalar()

    def update_nick(self, contact_pk: int, nick: Optional[str]) -> None:
        with self.session() as session:
            session.execute(
                update(Contact).where(Contact.id == contact_pk).values(nick=nick)
            )
            session.commit()

    def get_presence(self, contact_pk: int) -> Optional[CachedPresence]:
        with self.session() as session:
            presence = session.execute(
                select(
                    Contact.last_seen,
                    Contact.ptype,
                    Contact.pstatus,
                    Contact.pshow,
                    Contact.cached_presence,
                ).where(Contact.id == contact_pk)
            ).first()
            if presence is None or not presence[-1]:
                return None
            return CachedPresence(*presence[:-1])

    def set_presence(self, contact_pk: int, presence: CachedPresence) -> None:
        with self.session() as session:
            session.execute(
                update(Contact)
                .where(Contact.id == contact_pk)
                .values(**presence._asdict(), cached_presence=True)
            )
            session.commit()

    def reset_presence(self, contact_pk: int):
        with self.session() as session:
            session.execute(
                update(Contact)
                .where(Contact.id == contact_pk)
                .values(
                    last_seen=None,
                    ptype=None,
                    pstatus=None,
                    pshow=None,
                    cached_presence=False,
                )
            )
            session.commit()

    def set_avatar(self, contact_pk: int, avatar_pk: Optional[int]):
        with self.session() as session:
            session.execute(
                update(Contact)
                .where(Contact.id == contact_pk)
                .values(avatar_id=avatar_pk)
            )
            session.commit()

    def get_avatar_legacy_id(self, contact_pk: int) -> Optional[str]:
        with self.session() as session:
            contact = session.execute(
                select(Contact).where(Contact.id == contact_pk)
            ).scalar()
            if contact is None or contact.avatar is None:
                return None
            return contact.avatar.legacy_id

    def update(self, contact: "LegacyContact", commit=True) -> int:
        with self.session() as session:
            if contact.contact_pk is None:
                if contact.cached_presence is not None:
                    presence_kwargs = contact.cached_presence._asdict()
                    presence_kwargs["cached_presence"] = True
                else:
                    presence_kwargs = {}
                row = Contact(
                    jid=contact.jid.bare,
                    legacy_id=str(contact.legacy_id),
                    user_account_id=contact.user_pk,
                    **presence_kwargs,
                )
            else:
                row = (
                    session.query(Contact)
                    .filter(Contact.id == contact.contact_pk)
                    .one()
                )
            row.nick = contact.name
            row.is_friend = contact.is_friend
            row.added_to_roster = contact.added_to_roster
            row.updated = True
            row.extra_attributes = contact.serialize_extra_attributes()
            row.caps_ver = contact._caps_ver
            session.add(row)
            if commit:
                session.commit()
            return row.id

    def add_to_sent(self, contact_pk: int, msg_id: str) -> None:
        with self.session() as session:
            new = ContactSent(contact_id=contact_pk, msg_id=msg_id)
            session.add(new)
            session.commit()

    def pop_sent_up_to(self, contact_pk: int, msg_id: str) -> list[str]:
        result = []
        to_del = []
        with self.session() as session:
            for row in session.execute(
                select(ContactSent)
                .where(ContactSent.contact_id == contact_pk)
                .order_by(ContactSent.id)
            ).scalars():
                to_del.append(row.id)
                result.append(row.msg_id)
                if row.msg_id == msg_id:
                    break
            for row_id in to_del:
                session.execute(delete(ContactSent).where(ContactSent.id == row_id))
        return result

    def set_friend(self, contact_pk: int, is_friend: bool) -> None:
        with self.session() as session:
            session.execute(
                update(Contact)
                .where(Contact.id == contact_pk)
                .values(is_friend=is_friend)
            )
            session.commit()

    def set_added_to_roster(self, contact_pk: int, value: bool) -> None:
        with self.session() as session:
            session.execute(
                update(Contact)
                .where(Contact.id == contact_pk)
                .values(added_to_roster=value)
            )
            session.commit()

    def delete(self, contact_pk: int) -> None:
        with self.session() as session:
            session.execute(delete(Contact).where(Contact.id == contact_pk))
            session.commit()


class MAMStore(EngineMixin):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        with self.session() as session:
            session.execute(
                update(ArchivedMessage).values(source=ArchivedMessageSource.BACKFILL)
            )
            session.commit()

    def nuke_older_than(self, days: int) -> None:
        with self.session() as session:
            session.execute(
                delete(ArchivedMessage).where(
                    ArchivedMessage.timestamp < datetime.now() - timedelta(days=days)
                )
            )
            session.commit()

    def add_message(
        self,
        room_pk: int,
        message: HistoryMessage,
        archive_only: bool,
        legacy_msg_id: str | None,
    ) -> None:
        with self.session() as session:
            source = (
                ArchivedMessageSource.BACKFILL
                if archive_only
                else ArchivedMessageSource.LIVE
            )
            existing = session.execute(
                select(ArchivedMessage)
                .where(ArchivedMessage.room_id == room_pk)
                .where(ArchivedMessage.stanza_id == message.id)
            ).scalar()
            if existing is not None:
                log.debug("Updating message %s in room %s", message.id, room_pk)
                existing.timestamp = message.when
                existing.stanza = str(message.stanza)
                existing.author_jid = message.stanza.get_from()
                existing.source = source
                existing.legacy_id = legacy_msg_id
                session.add(existing)
                session.commit()
                return
            mam_msg = ArchivedMessage(
                stanza_id=message.id,
                timestamp=message.when,
                stanza=str(message.stanza),
                author_jid=message.stanza.get_from(),
                room_id=room_pk,
                source=source,
                legacy_id=legacy_msg_id,
            )
            session.add(mam_msg)
            session.commit()

    def get_messages(
        self,
        room_pk: int,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        ids: Collection[str] = (),
        last_page_n: Optional[int] = None,
        sender: Optional[str] = None,
        flip=False,
    ) -> Iterator[HistoryMessage]:

        with self.session() as session:
            q = select(ArchivedMessage).where(ArchivedMessage.room_id == room_pk)
            if start_date is not None:
                q = q.where(ArchivedMessage.timestamp >= start_date)
            if end_date is not None:
                q = q.where(ArchivedMessage.timestamp <= end_date)
            if before_id is not None:
                stamp = session.execute(
                    select(ArchivedMessage.timestamp).where(
                        ArchivedMessage.stanza_id == before_id
                    )
                ).scalar()
                if stamp is None:
                    raise XMPPError(
                        "item-not-found",
                        f"Message {before_id} not found",
                    )
                q = q.where(ArchivedMessage.timestamp < stamp)
            if after_id is not None:
                stamp = session.execute(
                    select(ArchivedMessage.timestamp).where(
                        ArchivedMessage.stanza_id == after_id
                    )
                ).scalar()
                if stamp is None:
                    raise XMPPError(
                        "item-not-found",
                        f"Message {after_id} not found",
                    )
                q = q.where(ArchivedMessage.timestamp > stamp)
            if ids:
                q = q.filter(ArchivedMessage.stanza_id.in_(ids))
            if sender is not None:
                q = q.where(ArchivedMessage.author_jid == sender)
            if flip:
                q = q.order_by(ArchivedMessage.timestamp.desc())
            else:
                q = q.order_by(ArchivedMessage.timestamp.asc())
            msgs = list(session.execute(q).scalars())
            if ids and len(msgs) != len(ids):
                raise XMPPError(
                    "item-not-found",
                    "One of the requested messages IDs could not be found "
                    "with the given constraints.",
                )
            if last_page_n is not None:
                if flip:
                    msgs = msgs[:last_page_n]
                else:
                    msgs = msgs[-last_page_n:]
            for h in msgs:
                yield HistoryMessage(
                    stanza=str(h.stanza), when=h.timestamp.replace(tzinfo=timezone.utc)
                )

    def get_first(self, room_pk: int, with_legacy_id=False) -> ArchivedMessage | None:
        with self.session() as session:
            q = (
                select(ArchivedMessage)
                .where(ArchivedMessage.room_id == room_pk)
                .order_by(ArchivedMessage.timestamp.asc())
            )
            if with_legacy_id:
                q = q.filter(ArchivedMessage.legacy_id.isnot(None))
            return session.execute(q).scalar()

    def get_last(
        self, room_pk: int, source: ArchivedMessageSource | None = None
    ) -> ArchivedMessage | None:
        with self.session() as session:
            q = select(ArchivedMessage).where(ArchivedMessage.room_id == room_pk)

            if source is not None:
                q = q.where(ArchivedMessage.source == source)

            return session.execute(
                q.order_by(ArchivedMessage.timestamp.desc())
            ).scalar()

    def get_first_and_last(self, room_pk: int) -> list[MamMetadata]:
        r = []
        with self.session():
            first = self.get_first(room_pk)
            if first is not None:
                r.append(MamMetadata(first.stanza_id, first.timestamp))
            last = self.get_last(room_pk)
            if last is not None:
                r.append(MamMetadata(last.stanza_id, last.timestamp))
        return r

    def get_most_recent_with_legacy_id(
        self, room_pk: int, source: ArchivedMessageSource | None = None
    ) -> ArchivedMessage | None:
        with self.session() as session:
            q = (
                select(ArchivedMessage)
                .where(ArchivedMessage.room_id == room_pk)
                .where(ArchivedMessage.legacy_id.isnot(None))
            )
            if source is not None:
                q = q.where(ArchivedMessage.source == source)
            return session.execute(
                q.order_by(ArchivedMessage.timestamp.desc())
            ).scalar()

    def get_least_recent_with_legacy_id_after(
        self, room_pk: int, after_id: str, source=ArchivedMessageSource.LIVE
    ) -> ArchivedMessage | None:
        with self.session() as session:
            after_timestamp = (
                session.query(ArchivedMessage.timestamp)
                .filter(ArchivedMessage.legacy_id == after_id)
                .scalar()
            )
            q = (
                select(ArchivedMessage)
                .where(ArchivedMessage.room_id == room_pk)
                .where(ArchivedMessage.legacy_id.isnot(None))
                .where(ArchivedMessage.source == source)
                .where(ArchivedMessage.timestamp > after_timestamp)
            )
            return session.execute(q.order_by(ArchivedMessage.timestamp.asc())).scalar()

    def get_by_legacy_id(self, room_pk: int, legacy_id: str) -> ArchivedMessage | None:
        with self.session() as session:
            return (
                session.query(ArchivedMessage)
                .filter(ArchivedMessage.room_id == room_pk)
                .filter(ArchivedMessage.legacy_id == legacy_id)
                .first()
            )


class MultiStore(EngineMixin):
    def get_xmpp_ids(self, user_pk: int, xmpp_id: str) -> list[str]:
        with self.session() as session:
            multi = session.execute(
                select(XmppIdsMulti)
                .where(XmppIdsMulti.xmpp_id == xmpp_id)
                .where(XmppIdsMulti.user_account_id == user_pk)
            ).scalar()
            if multi is None:
                return []
            return [m.xmpp_id for m in multi.legacy_ids_multi.xmpp_ids]

    def set_xmpp_ids(
        self, user_pk: int, legacy_msg_id: str, xmpp_ids: list[str], fail=False
    ) -> None:
        with self.session() as session:
            existing = session.execute(
                select(LegacyIdsMulti)
                .where(LegacyIdsMulti.user_account_id == user_pk)
                .where(LegacyIdsMulti.legacy_id == legacy_msg_id)
            ).scalar()
            if existing is not None:
                if fail:
                    raise
                log.warning("Resetting multi for %s", legacy_msg_id)
                session.execute(
                    delete(LegacyIdsMulti)
                    .where(LegacyIdsMulti.user_account_id == user_pk)
                    .where(LegacyIdsMulti.legacy_id == legacy_msg_id)
                )
                for i in xmpp_ids:
                    session.execute(
                        delete(XmppIdsMulti)
                        .where(XmppIdsMulti.user_account_id == user_pk)
                        .where(XmppIdsMulti.xmpp_id == i)
                    )
                session.commit()
                self.set_xmpp_ids(user_pk, legacy_msg_id, xmpp_ids, True)
                return

            row = LegacyIdsMulti(
                user_account_id=user_pk,
                legacy_id=legacy_msg_id,
                xmpp_ids=[
                    XmppIdsMulti(user_account_id=user_pk, xmpp_id=i)
                    for i in xmpp_ids
                    if i
                ],
            )
            session.add(row)
            session.commit()

    def get_legacy_id(self, user_pk: int, xmpp_id: str) -> Optional[str]:
        with self.session() as session:
            multi = session.execute(
                select(XmppIdsMulti)
                .where(XmppIdsMulti.xmpp_id == xmpp_id)
                .where(XmppIdsMulti.user_account_id == user_pk)
            ).scalar()
            if multi is None:
                return None
            return multi.legacy_ids_multi.legacy_id


class AttachmentStore(EngineMixin):
    def get_url(self, legacy_file_id: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(Attachment.url).where(
                    Attachment.legacy_file_id == legacy_file_id
                )
            ).scalar()

    def set_url(self, user_pk: int, legacy_file_id: str, url: str) -> None:
        with self.session() as session:
            att = session.execute(
                select(Attachment)
                .where(Attachment.legacy_file_id == legacy_file_id)
                .where(Attachment.user_account_id == user_pk)
            ).scalar()
            if att is None:
                att = Attachment(
                    legacy_file_id=legacy_file_id, url=url, user_account_id=user_pk
                )
                session.add(att)
            else:
                att.url = url
            session.commit()

    def get_sims(self, url: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(Attachment.sims).where(Attachment.url == url)
            ).scalar()

    def set_sims(self, url: str, sims: str) -> None:
        with self.session() as session:
            session.execute(
                update(Attachment).where(Attachment.url == url).values(sims=sims)
            )
            session.commit()

    def get_sfs(self, url: str) -> Optional[str]:
        with self.session() as session:
            return session.execute(
                select(Attachment.sfs).where(Attachment.url == url)
            ).scalar()

    def set_sfs(self, url: str, sfs: str) -> None:
        with self.session() as session:
            session.execute(
                update(Attachment).where(Attachment.url == url).values(sfs=sfs)
            )
            session.commit()

    def remove(self, legacy_file_id: str) -> None:
        with self.session() as session:
            session.execute(
                delete(Attachment).where(Attachment.legacy_file_id == legacy_file_id)
            )
            session.commit()


class RoomStore(UpdatedMixin):
    model = Room

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        with self.session() as session:
            session.execute(
                update(Room).values(
                    subject_setter=None, user_resources=None, history_filled=False
                )
            )
            session.commit()

    def set_avatar(self, room_pk: int, avatar_pk: int | None) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(avatar_id=avatar_pk)
            )
            session.commit()

    def get_avatar_legacy_id(self, room_pk: int) -> Optional[str]:
        with self.session() as session:
            room = session.execute(select(Room).where(Room.id == room_pk)).scalar()
            if room is None or room.avatar is None:
                return None
            return room.avatar.legacy_id

    def get_by_jid(self, user_pk: int, jid: JID) -> Optional[Room]:
        if jid.resource:
            raise TypeError
        with self.session() as session:
            return session.execute(
                select(Room)
                .where(Room.user_account_id == user_pk)
                .where(Room.jid == jid)
            ).scalar()

    def get_by_legacy_id(self, user_pk: int, legacy_id: str) -> Optional[Room]:
        with self.session() as session:
            return session.execute(
                select(Room)
                .where(Room.user_account_id == user_pk)
                .where(Room.legacy_id == legacy_id)
            ).scalar()

    def update_subject_setter(self, room_pk: int, subject_setter: str | None):
        with self.session() as session:
            session.execute(
                update(Room)
                .where(Room.id == room_pk)
                .values(subject_setter=subject_setter)
            )
            session.commit()

    def update(self, muc: "LegacyMUC") -> int:
        with self.session() as session:
            if muc.pk is None:
                row = Room(
                    jid=muc.jid,
                    legacy_id=str(muc.legacy_id),
                    user_account_id=muc.user_pk,
                )
            else:
                row = session.query(Room).filter(Room.id == muc.pk).one()

            row.updated = True
            row.extra_attributes = muc.serialize_extra_attributes()
            row.name = muc.name
            row.description = muc.description
            row.user_resources = (
                None
                if not muc._user_resources
                else json.dumps(list(muc._user_resources))
            )
            row.muc_type = muc.type
            row.subject = muc.subject
            row.subject_date = muc.subject_date
            row.subject_setter = muc.subject_setter
            row.participants_filled = muc._participants_filled
            row.n_participants = muc._n_participants
            row.user_nick = muc.user_nick
            session.add(row)
            session.commit()
            return row.id

    def update_subject_date(
        self, room_pk: int, subject_date: Optional[datetime]
    ) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(subject_date=subject_date)
            )
            session.commit()

    def update_subject(self, room_pk: int, subject: Optional[str]) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(subject=subject)
            )
            session.commit()

    def update_description(self, room_pk: int, desc: Optional[str]) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(description=desc)
            )
            session.commit()

    def update_name(self, room_pk: int, name: Optional[str]) -> None:
        with self.session() as session:
            session.execute(update(Room).where(Room.id == room_pk).values(name=name))
            session.commit()

    def update_n_participants(self, room_pk: int, n: Optional[int]) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(n_participants=n)
            )
            session.commit()

    def delete(self, room_pk: int) -> None:
        with self.session() as session:
            session.execute(delete(Room).where(Room.id == room_pk))
            session.execute(delete(Participant).where(Participant.room_id == room_pk))
            session.commit()

    def set_resource(self, room_pk: int, resources: set[str]) -> None:
        with self.session() as session:
            session.execute(
                update(Room)
                .where(Room.id == room_pk)
                .values(
                    user_resources=(
                        None if not resources else json.dumps(list(resources))
                    )
                )
            )
            session.commit()

    def nickname_is_available(self, room_pk: int, nickname: str) -> bool:
        with self.session() as session:
            return (
                session.execute(
                    select(Participant)
                    .where(Participant.room_id == room_pk)
                    .where(Participant.nickname == nickname)
                ).scalar()
                is None
            )

    def set_participants_filled(self, room_pk: int, val=True) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(participants_filled=val)
            )
            session.commit()

    def set_history_filled(self, room_pk: int, val=True) -> None:
        with self.session() as session:
            session.execute(
                update(Room).where(Room.id == room_pk).values(history_filled=True)
            )
            session.commit()

    def get_all(self, user_pk: int) -> Iterator[Room]:
        with self.session() as session:
            yield from session.execute(
                select(Room).where(Room.user_account_id == user_pk)
            ).scalars()


class ParticipantStore(EngineMixin):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        with self.session() as session:
            session.execute(delete(Participant))
            session.execute(delete(Hat))
            session.commit()

    def add(self, room_pk: int, nickname: str) -> int:
        with self.session() as session:
            existing = session.execute(
                select(Participant.id)
                .where(Participant.room_id == room_pk)
                .where(Participant.nickname == nickname)
            ).scalar()
            if existing is not None:
                return existing
            participant = Participant(room_id=room_pk, nickname=nickname)
            session.add(participant)
            session.commit()
            return participant.id

    def get_by_nickname(self, room_pk: int, nickname: str) -> Optional[Participant]:
        with self.session() as session:
            return session.execute(
                select(Participant)
                .where(Participant.room_id == room_pk)
                .where(Participant.nickname == nickname)
            ).scalar()

    def get_by_resource(self, room_pk: int, resource: str) -> Optional[Participant]:
        with self.session() as session:
            return session.execute(
                select(Participant)
                .where(Participant.room_id == room_pk)
                .where(Participant.resource == resource)
            ).scalar()

    def get_by_contact(self, room_pk: int, contact_pk: int) -> Optional[Participant]:
        with self.session() as session:
            return session.execute(
                select(Participant)
                .where(Participant.room_id == room_pk)
                .where(Participant.contact_id == contact_pk)
            ).scalar()

    def get_all(self, room_pk: int, user_included=True) -> Iterator[Participant]:
        with self.session() as session:
            q = select(Participant).where(Participant.room_id == room_pk)
            if not user_included:
                q = q.where(~Participant.is_user)
            yield from session.execute(q).scalars()

    def get_for_contact(self, contact_pk: int) -> Iterator[Participant]:
        with self.session() as session:
            yield from session.execute(
                select(Participant).where(Participant.contact_id == contact_pk)
            ).scalars()

    def update(self, participant: "LegacyParticipant") -> None:
        with self.session() as session:
            session.execute(
                update(Participant)
                .where(Participant.id == participant.pk)
                .values(
                    resource=participant.jid.resource,
                    affiliation=participant.affiliation,
                    role=participant.role,
                    presence_sent=participant._presence_sent,  # type:ignore
                    # hats=[self.add_hat(h.uri, h.title) for h in participant._hats],
                    is_user=participant.is_user,
                    contact_id=(
                        None
                        if participant.contact is None
                        else participant.contact.contact_pk
                    ),
                )
            )
            session.commit()

    def add_hat(self, uri: str, title: str) -> Hat:
        with self.session() as session:
            existing = session.execute(
                select(Hat).where(Hat.uri == uri).where(Hat.title == title)
            ).scalar()
            if existing is not None:
                return existing
            hat = Hat(uri=uri, title=title)
            session.add(hat)
            session.commit()
            return hat

    def set_presence_sent(self, participant_pk: int) -> None:
        with self.session() as session:
            session.execute(
                update(Participant)
                .where(Participant.id == participant_pk)
                .values(presence_sent=True)
            )
            session.commit()

    def set_affiliation(self, participant_pk: int, affiliation: MucAffiliation) -> None:
        with self.session() as session:
            session.execute(
                update(Participant)
                .where(Participant.id == participant_pk)
                .values(affiliation=affiliation)
            )
            session.commit()

    def set_role(self, participant_pk: int, role: MucRole) -> None:
        with self.session() as session:
            session.execute(
                update(Participant)
                .where(Participant.id == participant_pk)
                .values(role=role)
            )
            session.commit()

    def set_hats(self, participant_pk: int, hats: list[HatTuple]) -> None:
        with self.session() as session:
            part = session.execute(
                select(Participant).where(Participant.id == participant_pk)
            ).scalar()
            if part is None:
                raise ValueError
            part.hats.clear()
            for h in hats:
                hat = self.add_hat(*h)
                if hat in part.hats:
                    continue
                part.hats.append(hat)
            session.commit()

    def delete(self, participant_pk: int) -> None:
        with self.session() as session:
            session.execute(delete(Participant).where(Participant.id == participant_pk))

    def get_count(self, room_pk: int) -> int:
        with self.session() as session:
            return session.query(
                count(Participant.id).filter(Participant.room_id == room_pk)
            ).scalar()


log = logging.getLogger(__name__)
_session: Optional[Session] = None



"""Store room subject setter by nickname

Revision ID: 2b1f45ab7379
Revises: c4a8ec35a0e8
Create Date: 2024-07-20 00:14:36.882689

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "2b1f45ab7379"
down_revision: Union[str, None] = "c4a8ec35a0e8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(sa.Column("subject_setter", sa.String(), nullable=True))
        batch_op.drop_constraint("subject_setter_id_foreign_key", type_="foreignkey")
        batch_op.drop_column("subject_setter_id")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(sa.Column("subject_setter_id", sa.INTEGER(), nullable=True))
        batch_op.create_foreign_key(
            "subject_setter_id_foreign_key",
            "participant",
            ["subject_setter_id"],
            ["id"],
        )
        batch_op.drop_column("subject_setter")
    # ### end Alembic commands ###



import asyncio
import logging
import warnings
from typing import TYPE_CHECKING, AsyncIterator, Generic, Iterator, Optional, Type

from slixmpp import JID
from slixmpp.exceptions import IqError, XMPPError
from slixmpp.jid import JID_UNESCAPE_TRANSFORMATIONS, _unescape_node

from ..core.mixins.lock import NamedLockMixin
from ..db.store import ContactStore
from ..util import SubclassableOnce
from ..util.types import LegacyContactType, LegacyUserIdType
from .contact import LegacyContact

if TYPE_CHECKING:
    from ..core.session import BaseSession


class ContactIsUser(Exception):
    pass


class LegacyRoster(
    Generic[LegacyUserIdType, LegacyContactType],
    NamedLockMixin,
    metaclass=SubclassableOnce,
):
    """
    Virtual roster of a gateway user, that allows to represent all
    of their contacts as singleton instances (if used properly and not too bugged).

    Every :class:`.BaseSession` instance will have its own :class:`.LegacyRoster` instance
    accessible via the :attr:`.BaseSession.contacts` attribute.

    Typically, you will mostly use the :meth:`.LegacyRoster.by_legacy_id` function to
    retrieve a contact instance.

    You might need to override :meth:`.LegacyRoster.legacy_id_to_jid_username` and/or
    :meth:`.LegacyRoster.jid_username_to_legacy_id` to incorporate some custom logic
    if you need some characters when translation JID user parts and legacy IDs.
    """

    def __init__(self, session: "BaseSession"):
        self._contact_cls: Type[LegacyContactType] = (
            LegacyContact.get_self_or_unique_subclass()
        )
        self._contact_cls.xmpp = session.xmpp
        self.__store: ContactStore = session.xmpp.store.contacts

        self.session = session
        self.log = logging.getLogger(f"{self.session.user_jid.bare}:roster")
        self.user_legacy_id: Optional[LegacyUserIdType] = None
        self.ready: asyncio.Future[bool] = self.session.xmpp.loop.create_future()
        self.__filling = False
        super().__init__()

    def __repr__(self):
        return f"<Roster of {self.session.user_jid}>"

    def __iter__(self) -> Iterator[LegacyContactType]:
        with self.__store.session():
            for stored in self.__store.get_all(user_pk=self.session.user_pk):
                yield self._contact_cls.from_store(self.session, stored)

    async def __finish_init_contact(
        self, legacy_id: LegacyUserIdType, jid_username: str, *args, **kwargs
    ):
        async with self.lock(("finish", legacy_id)):
            with self.__store.session():
                stored = self.__store.get_by_legacy_id(
                    self.session.user_pk, str(legacy_id)
                )
                if stored is not None:
                    if stored.updated:
                        return self._contact_cls.from_store(self.session, stored)
                    c: LegacyContact = self._contact_cls(
                        self.session, legacy_id, jid_username, *args, **kwargs
                    )
                    c.contact_pk = stored.id
                else:
                    c = self._contact_cls(
                        self.session, legacy_id, jid_username, *args, **kwargs
                    )
                try:
                    with c.updating_info():
                        await c.avatar_wrap_update_info()
                except Exception as e:
                    raise XMPPError("internal-server-error", str(e))
                c._caps_ver = await c.get_caps_ver(c.jid)
                need_avatar = c.contact_pk is None
                c.contact_pk = self.__store.update(c, commit=not self.__filling)
                if need_avatar:
                    c._post_avatar_update()
        return c

    def known_contacts(self, only_friends=True) -> dict[str, LegacyContactType]:
        if only_friends:
            return {c.jid.bare: c for c in self if c.is_friend}
        return {c.jid.bare: c for c in self}

    async def by_jid(self, contact_jid: JID) -> LegacyContactType:
        # """
        # Retrieve a contact by their JID
        #
        # If the contact was not instantiated before, it will be created
        # using :meth:`slidge.LegacyRoster.jid_username_to_legacy_id` to infer their
        # legacy user ID.
        #
        # :param contact_jid:
        # :return:
        # """
        username = contact_jid.node
        async with self.lock(("username", username)):
            with self.__store.session():
                stored = self.__store.get_by_jid(self.session.user_pk, contact_jid)
                if stored is not None and stored.updated:
                    return self._contact_cls.from_store(self.session, stored)

            legacy_id = await self.jid_username_to_legacy_id(username)
            log.debug("Contact %s not found", contact_jid)
            if self.get_lock(("legacy_id", legacy_id)):
                log.debug("Already updating %s", contact_jid)
                return await self.by_legacy_id(legacy_id)
            return await self.__finish_init_contact(legacy_id, username)

    async def by_legacy_id(
        self, legacy_id: LegacyUserIdType, *args, **kwargs
    ) -> LegacyContactType:
        """
        Retrieve a contact by their legacy_id

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.legacy_id_to_jid_username` to infer their
        legacy user ID.

        :param legacy_id:
        :param args: arbitrary additional positional arguments passed to the contact constructor.
            Requires subclassing LegacyContact.__init__ to accept those.
            This is useful for networks where you fetch the contact list and information
            about these contacts in a single request
        :param kwargs: arbitrary keyword arguments passed to the contact constructor
        :return:
        """
        if legacy_id == self.user_legacy_id:
            raise ContactIsUser
        async with self.lock(("legacy_id", legacy_id)):
            with self.__store.session():
                stored = self.__store.get_by_legacy_id(
                    self.session.user_pk, str(legacy_id)
                )
                if stored is not None and stored.updated:
                    return self._contact_cls.from_store(
                        self.session, stored, *args, **kwargs
                    )

            username = await self.legacy_id_to_jid_username(legacy_id)
            log.debug("Contact %s not found", legacy_id)
            if self.get_lock(("username", username)):
                log.debug("Already updating %s", username)
                jid = JID()
                jid.node = username
                jid.domain = self.session.xmpp.boundjid.bare
                return await self.by_jid(jid)
            return await self.__finish_init_contact(
                legacy_id, username, *args, **kwargs
            )

    async def by_stanza(self, s) -> LegacyContact:
        # """
        # Retrieve a contact by the destination of a stanza
        #
        # See :meth:`slidge.Roster.by_legacy_id` for more info.
        #
        # :param s:
        # :return:
        # """
        return await self.by_jid(s.get_to())

    async def legacy_id_to_jid_username(self, legacy_id: LegacyUserIdType) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it is case-sensitive or contains
        forbidden characters not covered by :xep:`0106`.

        :param legacy_id:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, jid_username: str) -> LegacyUserIdType:
        """
        Convert a JID user part to a legacy ID.

        Should be overridden in case legacy IDs are not strings, or more generally
        for any case where the username part of a JID (unescaped with to the mapping
        defined by :xep:`0106`) is not enough to identify a contact on the legacy network.

        Default implementation is an identity operation

        :param jid_username: User part of a JID, ie "user" in "user@example.com"
        :return: An identifier for the user on the legacy network.
        """
        return _unescape_node(jid_username)

    async def _fill(self):
        try:
            if hasattr(self.session.xmpp, "TEST_MODE"):
                # dirty hack to avoid mocking xmpp server replies to this
                # during tests
                raise PermissionError
            iq = await self.session.xmpp["xep_0356"].get_roster(
                self.session.user_jid.bare
            )
            user_roster = iq["roster"]["items"]
        except (PermissionError, IqError):
            user_roster = None

        with self.__store.session() as orm:
            self.__filling = True
            async for contact in self.fill():
                if user_roster is None:
                    continue
                item = contact.get_roster_item()
                old = user_roster.get(contact.jid.bare)
                if old is not None and all(
                    old[k] == item[contact.jid.bare][k]
                    for k in ("subscription", "groups", "name")
                ):
                    self.log.debug("No need to update roster")
                    continue
                self.log.debug("Updating roster")
                try:
                    await self.session.xmpp["xep_0356"].set_roster(
                        self.session.user_jid.bare,
                        item,
                    )
                except (PermissionError, IqError) as e:
                    warnings.warn(f"Could not add to roster: {e}")
                else:
                    contact._added_to_roster = True
            orm.commit()
        self.__filling = False

    async def fill(self) -> AsyncIterator[LegacyContact]:
        """
        Populate slidge's "virtual roster".

        This should yield contacts that are meant to be added to the user's
        roster, typically by using ``await self.by_legacy_id(contact_id)``.
        Setting the contact nicknames, avatar, etc. should be in
        :meth:`LegacyContact.update_info()`

        It's not mandatory to override this method, but it is recommended way
        to populate "friends" of the user. Calling
        ``await (await self.by_legacy_id(contact_id)).add_to_roster()``
        accomplishes the same thing, but doing it in here allows to batch
        DB queries and is better performance-wise.

        """
        return
        yield


ESCAPE_TABLE = "".maketrans({v: k for k, v in JID_UNESCAPE_TRANSFORMATIONS.items()})
log = logging.getLogger(__name__)



"""
Mixins
"""

from typing import Optional

from .avatar import AvatarMixin
from .disco import ChatterDiscoMixin
from .message import MessageCarbonMixin, MessageMixin
from .presence import PresenceMixin


class FullMixin(ChatterDiscoMixin, MessageMixin, PresenceMixin):
    pass


class FullCarbonMixin(ChatterDiscoMixin, MessageCarbonMixin, PresenceMixin):
    pass


class StoredAttributeMixin:
    def serialize_extra_attributes(self) -> Optional[dict]:
        return None

    def deserialize_extra_attributes(self, data: dict) -> None:
        pass


__all__ = ("AvatarMixin", "FullCarbonMixin", "StoredAttributeMixin")



"""Rely on DB to store contacts, rooms and participants

Revision ID: 8d2ced764698
Revises: b33993e87db3
Create Date: 2024-07-08 14:39:47.022088

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

import slidge.db.meta

# revision identifiers, used by Alembic.
revision: str = "8d2ced764698"
down_revision: Union[str, None] = "b33993e87db3"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "hat",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("title", sa.String(), nullable=False),
        sa.Column("uri", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("title", "uri"),
    )
    op.create_table(
        "contact_sent",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("contact_id", sa.Integer(), nullable=False),
        sa.Column("msg_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["contact_id"],
            ["contact.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("contact_id", "msg_id"),
    )
    op.create_table(
        "participant",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("room_id", sa.Integer(), nullable=False),
        sa.Column("contact_id", sa.Integer(), nullable=True),
        sa.Column("is_user", sa.Boolean(), nullable=False),
        sa.Column(
            "affiliation",
            sa.Enum("outcast", "member", "admin", "owner", "none", native_enum=False),
            nullable=False,
        ),
        sa.Column(
            "role",
            sa.Enum("moderator", "participant", "visitor", "none", native_enum=False),
            nullable=False,
        ),
        sa.Column("presence_sent", sa.Boolean(), nullable=False),
        sa.Column("resource", sa.String(), nullable=True),
        sa.Column("nickname", sa.String(), nullable=True),
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
        sa.ForeignKeyConstraint(
            ["contact_id"],
            ["contact.id"],
        ),
        sa.ForeignKeyConstraint(
            ["room_id"],
            ["room.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "participant_hats",
        sa.Column("participant_id", sa.Integer(), nullable=False),
        sa.Column("hat_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["hat_id"],
            ["hat.id"],
        ),
        sa.ForeignKeyConstraint(
            ["participant_id"],
            ["participant.id"],
        ),
        sa.PrimaryKeyConstraint("participant_id", "hat_id"),
    )
    op.add_column("contact", sa.Column("is_friend", sa.Boolean(), nullable=False))
    op.add_column("contact", sa.Column("added_to_roster", sa.Boolean(), nullable=False))
    op.add_column(
        "contact",
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
    )
    op.add_column("contact", sa.Column("updated", sa.Boolean(), nullable=False))
    op.add_column("room", sa.Column("description", sa.String(), nullable=True))
    op.add_column("room", sa.Column("subject", sa.String(), nullable=True))
    op.add_column("room", sa.Column("subject_date", sa.DateTime(), nullable=True))
    op.add_column(
        "room",
        sa.Column(
            "muc_type",
            sa.Enum("GROUP", "CHANNEL", "CHANNEL_NON_ANONYMOUS", name="muctype"),
            nullable=True,
        ),
    )
    op.add_column("room", sa.Column("user_resources", sa.String(), nullable=True))
    op.add_column(
        "room", sa.Column("participants_filled", sa.Boolean(), nullable=False)
    )
    op.add_column(
        "room",
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
    )
    op.add_column("room", sa.Column("updated", sa.Boolean(), nullable=False))


def downgrade() -> None:
    op.drop_column("room", "updated")
    op.drop_column("room", "extra_attributes")
    op.drop_column("room", "participants_filled")
    op.drop_column("room", "user_resources")
    op.drop_column("room", "muc_type")
    op.drop_column("room", "subject_date")
    op.drop_column("room", "subject")
    op.drop_column("room", "description")
    op.drop_column("contact", "updated")
    op.drop_column("contact", "extra_attributes")
    op.drop_column("contact", "added_to_roster")
    op.drop_column("contact", "is_friend")
    op.drop_table("participant_hats")
    op.drop_table("participant")
    op.drop_table("contact_sent")
    op.drop_table("hat")



"""Store contacts caps verstring in DB

Revision ID: 2461390c0af2
Revises: 2b1f45ab7379
Create Date: 2024-07-20 08:00:11.675735

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "2461390c0af2"
down_revision: Union[str, None] = "2b1f45ab7379"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("contact", schema=None) as batch_op:
        batch_op.add_column(sa.Column("caps_ver_bare", sa.String(), nullable=True))
        batch_op.add_column(sa.Column("caps_ver", sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("contact", schema=None) as batch_op:
        batch_op.drop_column("caps_ver")
        batch_op.drop_column("caps_ver_bare")

    # ### end Alembic commands ###



import abc
import logging
from typing import TYPE_CHECKING, Generic, Iterator, Optional, Type

from slixmpp import JID
from slixmpp.exceptions import XMPPError
from slixmpp.jid import _unescape_node

from ..contact.roster import ESCAPE_TABLE
from ..core.mixins.lock import NamedLockMixin
from ..util import SubclassableOnce
from ..util.types import LegacyGroupIdType, LegacyMUCType
from .archive import MessageArchive
from .room import LegacyMUC

if TYPE_CHECKING:
    from slidge.core.session import BaseSession


class LegacyBookmarks(
    Generic[LegacyGroupIdType, LegacyMUCType],
    NamedLockMixin,
    metaclass=SubclassableOnce,
):
    """
    This is instantiated once per :class:`~slidge.BaseSession`
    """

    def __init__(self, session: "BaseSession"):
        self.session = session
        self.xmpp = session.xmpp
        self.user_jid = session.user_jid
        self.__store = self.xmpp.store.rooms

        self._muc_class: Type[LegacyMUCType] = LegacyMUC.get_self_or_unique_subclass()

        self._user_nick: str = self.session.user_jid.node

        super().__init__()
        self.log = logging.getLogger(f"{self.user_jid.bare}:bookmarks")
        self.ready = self.session.xmpp.loop.create_future()
        if not self.xmpp.GROUPS:
            self.ready.set_result(True)

    @property
    def user_nick(self):
        return self._user_nick

    @user_nick.setter
    def user_nick(self, nick: str):
        self._user_nick = nick

    def __iter__(self) -> Iterator[LegacyMUCType]:
        for stored in self.__store.get_all(user_pk=self.session.user_pk):
            yield self._muc_class.from_store(self.session, stored)

    def __repr__(self):
        return f"<Bookmarks of {self.user_jid}>"

    async def __finish_init_muc(self, legacy_id: LegacyGroupIdType, jid: JID):
        with self.__store.session():
            stored = self.__store.get_by_legacy_id(self.session.user_pk, str(legacy_id))
            if stored is not None:
                if stored.updated:
                    return self._muc_class.from_store(self.session, stored)
                muc = self._muc_class(self.session, legacy_id=legacy_id, jid=jid)
                muc.pk = stored.id
            else:
                muc = self._muc_class(self.session, legacy_id=legacy_id, jid=jid)

            try:
                with muc.updating_info():
                    await muc.avatar_wrap_update_info()
            except Exception as e:
                raise XMPPError("internal-server-error", str(e))
            if not muc.user_nick:
                muc.user_nick = self._user_nick
            self.log.debug("MUC created: %r", muc)
            muc.pk = self.__store.update(muc)
            muc.archive = MessageArchive(muc.pk, self.xmpp.store.mam)
        return muc

    async def legacy_id_to_jid_local_part(self, legacy_id: LegacyGroupIdType):
        return await self.legacy_id_to_jid_username(legacy_id)

    async def jid_local_part_to_legacy_id(self, local_part: str):
        return await self.jid_username_to_legacy_id(local_part)

    async def legacy_id_to_jid_username(self, legacy_id: LegacyGroupIdType):
        """
        The default implementation calls ``str()`` on the legacy_id and
        escape characters according to :xep:`0106`.

        You can override this class and implement a more subtle logic to raise
        an :class:`~slixmpp.exceptions.XMPPError` early

        :param legacy_id:
        :return:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, username: str):
        """

        :param username:
        :return:
        """
        return _unescape_node(username)

    async def by_jid(self, jid: JID) -> LegacyMUCType:
        if jid.resource:
            jid = JID(jid.bare)
        bare = jid.bare
        async with self.lock(("bare", bare)):
            assert isinstance(jid.username, str)
            legacy_id = await self.jid_local_part_to_legacy_id(jid.username)
            if self.get_lock(("legacy_id", legacy_id)):
                self.log.debug("Not instantiating %s after all", jid)
                return await self.by_legacy_id(legacy_id)

            with self.__store.session():
                stored = self.__store.get_by_jid(self.session.user_pk, jid)
                if stored is not None and stored.updated:
                    return self._muc_class.from_store(self.session, stored)

            self.log.debug("Attempting to instantiate a new MUC for JID %s", jid)
            local_part = jid.node

            self.log.debug("%r is group %r", local_part, legacy_id)
            return await self.__finish_init_muc(legacy_id, JID(bare))

    def by_jid_only_if_exists(self, jid: JID) -> Optional[LegacyMUCType]:
        with self.__store.session():
            stored = self.__store.get_by_jid(self.session.user_pk, jid)
            if stored is not None and stored.updated:
                return self._muc_class.from_store(self.session, stored)
        return None

    async def by_legacy_id(self, legacy_id: LegacyGroupIdType) -> LegacyMUCType:
        async with self.lock(("legacy_id", legacy_id)):
            with self.__store.session():
                stored = self.__store.get_by_legacy_id(
                    self.session.user_pk, str(legacy_id)
                )
                if stored is not None and stored.updated:
                    return self._muc_class.from_store(self.session, stored)
            self.log.debug("Create new MUC instance for legacy ID %s", legacy_id)
            local = await self.legacy_id_to_jid_local_part(legacy_id)
            bare = f"{local}@{self.xmpp.boundjid}"
            jid = JID(bare)
            if self.get_lock(("bare", bare)):
                self.log.debug("Not instantiating %s after all", legacy_id)
                return await self.by_jid(jid)
            muc = await self.__finish_init_muc(legacy_id, jid)

            return muc

    @abc.abstractmethod
    async def fill(self):
        """
        Establish a user's known groups.

        This has to be overridden in plugins with group support and at the
        minimum, this should ``await self.by_legacy_id(group_id)`` for all
        the groups a user is part of.

        Slidge internals will call this on successful :meth:`BaseSession.login`

        """
        if self.xmpp.GROUPS:
            raise NotImplementedError(
                "The plugin advertised support for groups but"
                " LegacyBookmarks.fill() was not overridden."
            )

    def remove(self, muc: LegacyMUC):
        assert muc.pk is not None
        self.__store.delete(muc.pk)



from contextlib import contextmanager


class UpdateInfoMixin:
    """
    This mixin just adds a context manager that prevents commiting to the DB
    on every attribute change.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._updating_info = False

    @contextmanager
    def updating_info(self):
        self._updating_info = True
        yield
        self._updating_info = False



"""Store users' avatars' hashes persistently

Revision ID: e91195719c2c
Revises: aa9d82a7f6ef
Create Date: 2024-06-01 14:14:51.984943

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e91195719c2c"
down_revision: Union[str, None] = "aa9d82a7f6ef"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column("user_account", sa.Column("avatar_hash", sa.String(), nullable=True))


def downgrade() -> None:
    op.drop_column("user_account", "avatar_hash")



"""Add MUC.history_filled

Also drop caps_ver_bare column that should never have been added.

Revision ID: 82a4af84b679
Revises: 2461390c0af2
Create Date: 2024-07-22 07:01:05.352737

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "82a4af84b679"
down_revision: Union[str, None] = "2461390c0af2"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("contact", schema=None) as batch_op:
        batch_op.drop_column("caps_ver_bare")

    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "history_filled",
                sa.Boolean(),
                nullable=False,
                server_default=sa.False_(),  # manually added
            )
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.drop_column("history_filled")

    with op.batch_alter_table("contact", schema=None) as batch_op:
        batch_op.add_column(sa.Column("caps_ver_bare", sa.VARCHAR(), nullable=True))

    # ### end Alembic commands ###



from __future__ import annotations

import json
from typing import Union

import sqlalchemy as sa
from slixmpp import JID


class JIDType(sa.TypeDecorator[JID]):
    """
    Custom SQLAlchemy type for JIDs
    """

    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(self, value: JID | None, dialect: sa.Dialect) -> str | None:
        if value is None:
            return value
        return str(value)

    def process_result_value(
        self, value: str | None, dialect: sa.Dialect
    ) -> JID | None:
        if value is None:
            return value
        return JID(value)


class JSONEncodedDict(sa.TypeDecorator):
    """
    Custom SQLAlchemy type for dictionaries stored as JSON

    Note that mutations of the dictionary are not detected by SQLAlchemy,
    which is why use ``attributes.flag_modified()`` in ``UserStore.update()``
    """

    impl = sa.VARCHAR

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


JSONSerializableTypes = Union[str, float, None, "JSONSerializable"]
JSONSerializable = dict[str, JSONSerializableTypes]


class Base(sa.orm.DeclarativeBase):
    type_annotation_map = {JSONSerializable: JSONEncodedDict, JID: JIDType}


def get_engine(path: str) -> sa.Engine:
    engine = sa.create_engine(path)
    return engine



"""Move everything to persistent DB

Revision ID: b33993e87db3
Revises: e91195719c2c
Create Date: 2024-06-25 16:09:36.663953

"""

import shutil
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

import slidge.db.meta
from slidge import global_config

# revision identifiers, used by Alembic.
revision: str = "b33993e87db3"
down_revision: Union[str, None] = "e91195719c2c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "avatar",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("filename", sa.String(), nullable=False),
        sa.Column("hash", sa.String(), nullable=False),
        sa.Column("height", sa.Integer(), nullable=False),
        sa.Column("width", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column("etag", sa.String(), nullable=True),
        sa.Column("last_modified", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("filename"),
        sa.UniqueConstraint("hash"),
        sa.UniqueConstraint("legacy_id"),
    )
    op.create_table(
        "attachment",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_file_id", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=False),
        sa.Column("sims", sa.String(), nullable=True),
        sa.Column("sfs", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_attachment_legacy_file_id"),
        "attachment",
        ["legacy_file_id"],
        unique=False,
    )
    op.create_index(op.f("ix_attachment_url"), "attachment", ["url"], unique=False)
    op.create_table(
        "contact",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column("avatar_id", sa.Integer(), nullable=True),
        sa.Column("nick", sa.String(), nullable=True),
        sa.Column("cached_presence", sa.Boolean(), nullable=False),
        sa.Column("last_seen", sa.DateTime(), nullable=True),
        sa.Column("ptype", sa.String(), nullable=True),
        sa.Column("pstatus", sa.String(), nullable=True),
        sa.Column("pshow", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["avatar_id"],
            ["avatar.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_account_id", "jid"),
        sa.UniqueConstraint("user_account_id", "legacy_id"),
    )
    op.create_table(
        "legacy_ids_multi",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "legacy_ids_multi_user_account_id_legacy_id",
        "legacy_ids_multi",
        ["user_account_id", "legacy_id"],
        unique=True,
    )
    op.create_table(
        "room",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column("avatar_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["avatar_id"],
            ["avatar.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("jid"),
        sa.UniqueConstraint("legacy_id"),
    )
    op.create_table(
        "xmpp_to_legacy_ids",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column(
            "type",
            sa.Enum("DM", "GROUP_CHAT", "THREAD", name="xmpptolegacyenum"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "xmpp_legacy",
        "xmpp_to_legacy_ids",
        ["user_account_id", "xmpp_id", "legacy_id"],
        unique=True,
    )
    op.create_table(
        "mam",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("room_id", sa.Integer(), nullable=False),
        sa.Column("stanza_id", sa.String(), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=False),
        sa.Column("author_jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column("stanza", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["room_id"],
            ["room.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("room_id", "stanza_id"),
    )
    op.create_table(
        "xmpp_ids_multi",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.Column("legacy_ids_multi_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["legacy_ids_multi_id"],
            ["legacy_ids_multi.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "legacy_ids_multi_user_account_id_xmpp_id",
        "xmpp_ids_multi",
        ["user_account_id", "xmpp_id"],
        unique=True,
    )

    try:
        shutil.rmtree(global_config.HOME_DIR / "slidge_avatars_v2")
    except (FileNotFoundError, AttributeError):
        pass

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "legacy_ids_multi_user_account_id_xmpp_id", table_name="xmpp_ids_multi"
    )
    op.drop_table("xmpp_ids_multi")
    op.drop_table("mam")
    op.drop_index("xmpp_legacy", table_name="xmpp_to_legacy_ids")
    op.drop_table("xmpp_to_legacy_ids")
    op.drop_table("room")
    op.drop_index(
        "legacy_ids_multi_user_account_id_legacy_id", table_name="legacy_ids_multi"
    )
    op.drop_table("legacy_ids_multi")
    op.drop_table("contact")
    op.drop_index(op.f("ix_attachment_url"), table_name="attachment")
    op.drop_index(op.f("ix_attachment_legacy_file_id"), table_name="attachment")
    op.drop_table("attachment")
    op.drop_table("avatar")
    # ### end Alembic commands ###



"""DB Creation

Including a migration from the user_store shelf

Revision ID: aa9d82a7f6ef
Revises:
Create Date: 2024-04-17 20:57:01.357041

"""

import logging
from datetime import datetime
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

import slidge.db.meta

# revision identifiers, used by Alembic.
revision: str = "aa9d82a7f6ef"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    accounts = op.create_table(
        "user_account",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column(
            "registration_date",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "legacy_module_data", slidge.db.meta.JSONEncodedDict(), nullable=False
        ),
        sa.Column("preferences", slidge.db.meta.JSONEncodedDict(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("jid"),
    )
    # ### end Alembic commands ###
    try:
        migrate_from_shelf(accounts)
    except Exception:
        downgrade()
        raise


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user_account")
    # ### end Alembic commands ###


def migrate_from_shelf(accounts: sa.Table) -> None:
    try:
        from slidge.util.db import user_store
    except ImportError:
        return
    try:
        users = list(user_store.get_all())
    except AttributeError:
        return
    logging.info("Migrating %s users from the deprecated user_store shelf", len(users))
    op.bulk_insert(
        accounts,
        [
            {
                "jid": user.jid,
                "registration_date": (
                    user.registration_date
                    if user.registration_date is not None
                    else datetime.now()
                ),
                "legacy_module_data": user.registration_form,
                "preferences": {},
            }
            for user in users
        ],
    )



"""Add source and legacy ID for archived messages

Revision ID: b64b1a793483
Revises: 82a4af84b679
Create Date: 2024-07-22 21:06:35.020569

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

from slidge.db.models import ArchivedMessage

# revision identifiers, used by Alembic.
revision: str = "b64b1a793483"
down_revision: Union[str, None] = "82a4af84b679"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # since we don't want source to be nullable, we drop all rows first.
    # This is what you get by using alpha versions!
    op.execute(sa.delete(ArchivedMessage))
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("mam", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "source",
                sa.Enum("LIVE", "BACKFILL", name="archivedmessagesource"),
                nullable=False,
            )
        )
        batch_op.add_column(sa.Column("legacy_id", sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mam", schema=None) as batch_op:
        batch_op.drop_column("legacy_id")
        batch_op.drop_column("source")

    # ### end Alembic commands ###



"""Add n_participants attributes to Room

Should have been part of another commit, but I messed up some rebase

Revision ID: 09f27f098baa
Revises: 29f5280c61aa
Create Date: 2024-07-11 10:54:21.155871

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "09f27f098baa"
down_revision: Union[str, None] = "29f5280c61aa"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(sa.Column("n_participants", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.drop_column("n_participants")

    # ### end Alembic commands ###



import logging
import shutil
import sys
from pathlib import Path

from alembic import command
from alembic.config import Config

from .core import config


def remove_avatar_cache_v1():
    old_dir = config.HOME_DIR / "slidge_avatars"
    if old_dir.exists():
        log.info("Avatar cache dir v1 found, clearing it.")
        shutil.rmtree(old_dir)


def get_alembic_cfg() -> Config:
    alembic_cfg = Config()
    alembic_cfg.set_section_option(
        "alembic",
        "script_location",
        str(Path(__file__).parent / "db" / "alembic"),
    )
    return alembic_cfg


def migrate():
    remove_avatar_cache_v1()
    command.upgrade(get_alembic_cfg(), "head")


def main():
    """
    Updates the (dev) database in ./dev/slidge.sqlite and generates a revision

    Usage: python -m slidge.migration "Revision message blah blah blah"
    """
    alembic_cfg = get_alembic_cfg()
    command.upgrade(alembic_cfg, "head")
    command.revision(alembic_cfg, sys.argv[1], autogenerate=True)


log = logging.getLogger(__name__)

if __name__ == "__main__":
    main()



import asyncio
import hashlib
import io
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from http import HTTPStatus
from pathlib import Path
from typing import Any, Callable, Optional

import aiohttp
from multidict import CIMultiDictProxy
from PIL.Image import Image
from PIL.Image import open as open_image
from sqlalchemy import select

from slidge.core import config
from slidge.db.models import Avatar
from slidge.db.store import AvatarStore
from slidge.util.types import URL, AvatarType, LegacyFileIdType


@dataclass
class CachedAvatar:
    pk: int
    filename: str
    hash: str
    height: int
    width: int
    root: Path
    etag: Optional[str] = None
    last_modified: Optional[str] = None

    @property
    def data(self):
        return self.path.read_bytes()

    @property
    def path(self):
        return self.root / self.filename

    @staticmethod
    def from_store(stored: Avatar, root_dir: Path):
        return CachedAvatar(
            pk=stored.id,
            filename=stored.filename,
            hash=stored.hash,
            height=stored.height,
            width=stored.width,
            etag=stored.etag,
            root=root_dir,
            last_modified=stored.last_modified,
        )


class NotModified(Exception):
    pass


class AvatarCache:
    dir: Path
    http: aiohttp.ClientSession
    store: AvatarStore
    legacy_avatar_type: Callable[[str], Any] = str

    def __init__(self):
        self._thread_pool = ThreadPoolExecutor(config.AVATAR_RESAMPLING_THREADS)

    def set_dir(self, path: Path):
        self.dir = path
        self.dir.mkdir(exist_ok=True)
        with self.store.session():
            for stored in self.store.get_all():
                avatar = CachedAvatar.from_store(stored, root_dir=path)
                if avatar.path.exists():
                    continue
                log.warning(
                    "Removing avatar %s from store because %s does not exist",
                    avatar.hash,
                    avatar.path,
                )
                self.store.delete_by_pk(stored.id)

    def close(self):
        self._thread_pool.shutdown(cancel_futures=True)

    def __get_http_headers(self, cached: Optional[CachedAvatar | Avatar]):
        headers = {}
        if cached and (self.dir / cached.filename).exists():
            if last_modified := cached.last_modified:
                headers["If-Modified-Since"] = last_modified
            if etag := cached.etag:
                headers["If-None-Match"] = etag
        return headers

    async def __download(
        self,
        url: str,
        headers: dict[str, str],
    ) -> tuple[Image, CIMultiDictProxy[str]]:
        async with self.http.get(url, headers=headers) as response:
            if response.status == HTTPStatus.NOT_MODIFIED:
                log.debug("Using avatar cache for %s", url)
                raise NotModified
            return (
                open_image(io.BytesIO(await response.read())),
                response.headers,
            )

    async def __is_modified(self, url, headers) -> bool:
        async with self.http.head(url, headers=headers) as response:
            return response.status != HTTPStatus.NOT_MODIFIED

    async def url_modified(self, url: URL) -> bool:
        cached = self.store.get_by_url(url)
        if cached is None:
            return True
        headers = self.__get_http_headers(cached)
        return await self.__is_modified(url, headers)

    def get(self, unique_id: LegacyFileIdType | URL) -> Optional[CachedAvatar]:
        if isinstance(unique_id, URL):
            stored = self.store.get_by_url(unique_id)
        else:
            stored = self.store.get_by_legacy_id(str(unique_id))
        if stored is None:
            return None
        return CachedAvatar.from_store(stored, self.dir)

    def get_by_pk(self, pk: int) -> CachedAvatar:
        stored = self.store.get_by_pk(pk)
        assert stored is not None
        return CachedAvatar.from_store(stored, self.dir)

    @staticmethod
    async def _get_image(avatar: AvatarType) -> Image:
        if isinstance(avatar, bytes):
            return open_image(io.BytesIO(avatar))
        elif isinstance(avatar, Path):
            return open_image(avatar)
        raise TypeError("Avatar must be bytes or a Path", avatar)

    async def convert_or_get(
        self,
        avatar: AvatarType,
        unique_id: Optional[LegacyFileIdType],
    ) -> CachedAvatar:
        if unique_id is not None:
            cached = self.get(str(unique_id))
            if cached is not None:
                return cached

        if isinstance(avatar, (URL, str)):
            if unique_id is None:
                with self.store.session():
                    stored = self.store.get_by_url(avatar)
                    try:
                        img, response_headers = await self.__download(
                            avatar, self.__get_http_headers(stored)
                        )
                    except NotModified:
                        assert stored is not None
                        return CachedAvatar.from_store(stored, self.dir)
            else:
                img, _ = await self.__download(avatar, {})
                response_headers = None
        else:
            img = await self._get_image(avatar)
            response_headers = None
        with self.store.session() as orm:
            stored = orm.execute(
                select(Avatar).where(Avatar.legacy_id == str(unique_id))
            ).scalar()
            if stored is not None and stored.url is None:
                return CachedAvatar.from_store(stored, self.dir)

            resize = (size := config.AVATAR_SIZE) and any(x > size for x in img.size)
            if resize:
                await asyncio.get_event_loop().run_in_executor(
                    self._thread_pool, img.thumbnail, (size, size)
                )
                log.debug("Resampled image to %s", img.size)

            filename = str(uuid.uuid1()) + ".png"
            file_path = self.dir / filename

            if (
                not resize
                and img.format == "PNG"
                and isinstance(unique_id, str)
                and (path := Path(unique_id))
                and path.exists()
            ):
                img_bytes = path.read_bytes()
            else:
                with io.BytesIO() as f:
                    img.save(f, format="PNG")
                    img_bytes = f.getvalue()

            with file_path.open("wb") as file:
                file.write(img_bytes)

            hash_ = hashlib.sha1(img_bytes).hexdigest()

            stored = orm.execute(select(Avatar).where(Avatar.hash == hash_)).scalar()

            if stored is not None:
                return CachedAvatar.from_store(stored, self.dir)

            stored = Avatar(
                filename=filename,
                hash=hash_,
                height=img.height,
                width=img.width,
                legacy_id=None if unique_id is None else str(unique_id),
                url=avatar if isinstance(avatar, (URL, str)) else None,
            )
            if response_headers:
                stored.etag = response_headers.get("etag")
                stored.last_modified = response_headers.get("last-modified")

            orm.add(stored)
            orm.commit()
            return CachedAvatar.from_store(stored, self.dir)


avatar_cache = AvatarCache()
log = logging.getLogger(__name__)
_download_lock = asyncio.Lock()

__all__ = (
    "CachedAvatar",
    "avatar_cache",
)



"""Store subject setter in Room

Revision ID: 29f5280c61aa
Revises: 8d2ced764698
Create Date: 2024-07-10 13:09:25.181594

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "29f5280c61aa"
down_revision: Union[str, None] = "8d2ced764698"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(sa.Column("subject_setter_id", sa.Integer(), nullable=True))
        # we give this constraint a name a workaround for
        # https://github.com/sqlalchemy/alembic/issues/1195
        batch_op.create_foreign_key(
            "subject_setter_id_foreign_key",
            "participant",
            ["subject_setter_id"],
            ["id"],
        )


def downgrade() -> None:
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.drop_constraint("subject_setter_id_foreign_key", type_="foreignkey")
        batch_op.drop_column("subject_setter_id")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from ...db import GatewayUser

if TYPE_CHECKING:
    from .base import BaseGateway


class Registration:
    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp
        xmpp["xep_0077"].api.register(
            self.xmpp.make_registration_form, "make_registration_form"
        )
        xmpp["xep_0077"].api.register(self._user_get, "user_get")
        xmpp["xep_0077"].api.register(self._user_validate, "user_validate")
        xmpp["xep_0077"].api.register(self._user_modify, "user_modify")
        # kept for slixmpp internal API compat
        # TODO: either fully use slixmpp internal API or rewrite registration without it at all
        xmpp["xep_0077"].api.register(lambda *a: None, "user_remove")

    def get_user(self, jid: JID) -> GatewayUser | None:
        return self.xmpp.store.users.get(jid)

    async def _user_get(
        self, _gateway_jid, _node, ifrom: JID, iq: Iq
    ) -> GatewayUser | None:
        if ifrom is None:
            ifrom = iq.get_from()
        return self.get_user(ifrom)

    async def _user_validate(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        xmpp = self.xmpp
        log.debug("User validate: %s", ifrom.bare)
        form_dict = {f.var: iq.get(f.var) for f in xmpp.REGISTRATION_FIELDS}
        xmpp.raise_if_not_allowed_jid(ifrom)
        legacy_module_data = await xmpp.user_prevalidate(ifrom, form_dict)
        if legacy_module_data is None:
            legacy_module_data = form_dict
        user = self.xmpp.store.users.new(
            jid=ifrom,
            legacy_module_data=legacy_module_data,  # type:ignore
        )
        log.info("New user: %s", user)

    async def _user_modify(
        self, _gateway_jid, _node, ifrom: JID, form_dict: dict[str, Optional[str]]
    ):
        await self.xmpp.user_prevalidate(ifrom, form_dict)
        log.debug("Modify user: %s", ifrom)
        user = self.xmpp.store.users.get(ifrom)
        if user is None:
            raise XMPPError("internal-server-error", "User not found")
        user.legacy_module_data.update(form_dict)
        self.xmpp.store.users.update(user)


log = logging.getLogger(__name__)
